@@ -23,13 +23,16 @@
 
 #include <config.h>
 
+#include <util/log.h>
+#include <util/glib.h>
+#include <util/misc.h>
+
 #include "jsapi-util.h"
 #include "context-jsapi.h"
 
-#include <util/log.h>
-#include <util/glib.h>
-
 #include <string.h>
+#include <jscntxt.h>
+
 
 typedef struct {
     GHashTable *dynamic_classes;
@@ -82,7 +85,7 @@ gjs_runtime_get_load_context(JSRuntime *runtime)
     return gjs_context_get_context(context);
 }
 
-static JSContext*
+JSContext*
 gjs_runtime_peek_load_context(JSRuntime *runtime)
 {
     GjsContext *context;
@@ -253,9 +256,10 @@ gjs_object_get_property(JSContext  *context,
  */
 gboolean
 gjs_object_require_property(JSContext       *context,
-                               JSObject        *obj,
-                               const char      *property_name,
-                               jsval           *value_p)
+                            JSObject        *obj,
+                            const char      *obj_description,
+                            const char      *property_name,
+                            jsval           *value_p)
 {
     jsval value;
 
@@ -272,9 +276,14 @@ gjs_object_require_property(JSContext       *context,
         /* remember gjs_throw() is a no-op if JS_GetProperty()
          * already set an exception
          */
-        gjs_throw(context,
-                     "No property '%s' in object %p (or its value was undefined)",
-                     property_name, obj);
+        if (obj_description)
+            gjs_throw(context,
+                      "No property '%s' in %s (or its value was undefined)",
+                      property_name, obj_description);
+        else
+            gjs_throw(context,
+                      "No property '%s' in object %p (or its value was undefined)",
+                      property_name, obj);
         return FALSE;
     }
 }
@@ -296,6 +305,7 @@ gjs_init_class_dynamic(JSContext      *context,
     jsval value;
     char *private_name;
     JSObject *prototype;
+    JSContext *load_context;
 
     if (clasp->name != NULL) {
         g_warning("Dynamic class should not have a name in the JSClass struct");
@@ -308,7 +318,7 @@ gjs_init_class_dynamic(JSContext      *context,
      * the class in whatever global object initialized the
      * class first, which is not desirable.
      */
-    context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
 
     /* JS_InitClass() wants to define the constructor in the global object, so
      * we give it a private and namespaced name... passing in the namespace
@@ -319,25 +329,25 @@ gjs_init_class_dynamic(JSContext      *context,
     private_name = g_strdup_printf("_private_%s_%s", ns_name, class_name);
 
     prototype = NULL;
-    if (gjs_object_get_property(context, JS_GetGlobalObject(context),
+    if (gjs_object_get_property(load_context, JS_GetGlobalObject(load_context),
                                    private_name, &value) &&
         JSVAL_IS_OBJECT(value)) {
         jsval proto_val;
 
         g_free(private_name); /* don't need it anymore */
 
-        if (!gjs_object_require_property(context, JSVAL_TO_OBJECT(value),
+        if (!gjs_object_require_property(load_context, JSVAL_TO_OBJECT(value), NULL,
                                             "prototype", &proto_val) ||
             !JSVAL_IS_OBJECT(proto_val)) {
-            gjs_throw(context, "prototype was not defined or not an object?");
-            return NULL;
+            gjs_throw(load_context, "prototype was not defined or not an object?");
+            goto error;
         }
         prototype = JSVAL_TO_OBJECT(proto_val);
     } else {
         DynamicJSClass *class_copy;
         RuntimeData *rd;
 
-        rd = get_data_from_context(context);
+        rd = get_data_from_context(load_context);
 
         class_copy = g_slice_new0(DynamicJSClass);
         class_copy->base = *clasp;
@@ -355,7 +365,7 @@ gjs_init_class_dynamic(JSContext      *context,
                   "Initializing dynamic class %s %p",
                   class_name, class_copy);
 
-        prototype = JS_InitClass(context, JS_GetGlobalObject(context),
+        prototype = JS_InitClass(load_context, JS_GetGlobalObject(load_context),
                                  parent_proto, &class_copy->base,
                                  constructor, nargs,
                                  ps, fs,
@@ -364,9 +374,9 @@ gjs_init_class_dynamic(JSContext      *context,
         /* Retrieve the property again so we can define it in
          * in_object
          */
-        if (!gjs_object_require_property(context, JS_GetGlobalObject(context),
+        if (!gjs_object_require_property(load_context, JS_GetGlobalObject(load_context), NULL,
                                             class_copy->base.name, &value))
-            return NULL;
+            goto error;
     }
     g_assert(value != JSVAL_VOID);
     g_assert(prototype != NULL);
@@ -374,14 +384,24 @@ gjs_init_class_dynamic(JSContext      *context,
     /* Now manually define our constructor with a sane name, in the
      * namespace object.
      */
-    if (!JS_DefineProperty(context, in_object,
+    if (!JS_DefineProperty(load_context, in_object,
                            class_name,
                            value,
                            NULL, NULL,
                            GJS_MODULE_PROP_FLAGS))
-        return NULL;
+        goto error;
 
     return prototype;
+
+ error:
+    /* Move the exception to the calling context from load context.
+     */
+    if (!gjs_move_exception(load_context, context)) {
+        /* set an exception since none was set */
+        gjs_throw(context, "No exception was set, but class initialize failed somehow");
+    }
+
+    return NULL;
 }
 
 gboolean
@@ -432,6 +452,38 @@ gjs_get_instance_private_dynamic(JSContext      *context,
     return JS_GetInstancePrivate(context, obj, obj_class, argv);
 }
 
+void*
+gjs_get_instance_private_dynamic_with_typecheck(JSContext      *context,
+                                                JSObject       *obj,
+                                                JSClass        *static_clasp,
+                                                jsval          *argv)
+{
+    RuntimeData *rd;
+    JSClass *obj_class;
+
+    if (static_clasp->name != NULL) {
+        g_warning("Dynamic class should not have a name in the JSClass struct");
+        return NULL;
+    }
+
+    obj_class = JS_GetClass(context, obj);
+    g_assert(obj_class != NULL);
+
+    rd = get_data_from_context(context);
+    g_assert(rd != NULL);
+
+    /* Check that it's safe to cast to DynamicJSClass */
+    if (g_hash_table_lookup(rd->dynamic_classes, obj_class) == NULL) {
+        return NULL;
+    }
+
+    if (static_clasp != ((DynamicJSClass*) obj_class)->static_class) {
+        return NULL;
+    }
+
+    return JS_GetInstancePrivate(context, obj, obj_class, argv);
+}
+
 JSObject*
 gjs_construct_object_dynamic(JSContext      *context,
                              JSObject       *proto,
@@ -440,22 +492,24 @@ gjs_construct_object_dynamic(JSContext      *context,
 {
     RuntimeData *rd;
     JSClass *proto_class;
+    JSContext *load_context;
+    JSObject *result;
 
     /* We replace the passed-in context and global object with our
      * runtime-global permanent load context. Otherwise, JS_ConstructObject
      * can't find the constructor in whatever random global object is set
      * on the passed-in context.
      */
-    context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
 
-    proto_class = JS_GetClass(context, proto);
+    proto_class = JS_GetClass(load_context, proto);
 
-    rd = get_data_from_context(context);
+    rd = get_data_from_context(load_context);
 
     /* Check that it's safe to cast to DynamicJSClass */
     if (g_hash_table_lookup(rd->dynamic_classes, proto_class) == NULL) {
-        gjs_throw(context, "Prototype is not for a dynamically-registered class");
-        return NULL;
+        gjs_throw(load_context, "Prototype is not for a dynamically-registered class");
+        goto error;
     }
 
     gjs_debug_lifecycle(GJS_DEBUG_GREPO,
@@ -463,9 +517,24 @@ gjs_construct_object_dynamic(JSContext      *context,
                         proto_class->name, proto_class, proto);
 
     if (argc > 0)
-        return JS_ConstructObjectWithArguments(context, proto_class, proto, NULL, argc, argv);
+        result = JS_ConstructObjectWithArguments(load_context, proto_class, proto, NULL, argc, argv);
     else
-        return JS_ConstructObject(context, proto_class, proto, NULL);
+        result = JS_ConstructObject(load_context, proto_class, proto, NULL);
+
+    if (!result)
+        goto error;
+
+    return result;
+
+ error:
+    /* Move the exception to the calling context from load context.
+     */
+    if (!gjs_move_exception(load_context, context)) {
+        /* set an exception since none was set */
+        gjs_throw(context, "No exception was set, but object construction failed somehow");
+    }
+
+    return NULL;
 }
 
 JSObject*
@@ -810,6 +879,14 @@ gjs_error_reporter(JSContext     *context,
 {
     const char *warning;
 
+    if (gjs_environment_variable_is_set("GJS_ABORT_ON_OOM") &&
+        report->flags == JSREPORT_ERROR &&
+        report->errorNumber == JSMSG_OUT_OF_MEMORY) {
+        g_error("GJS ran out of memory at %s: %i.",
+                report->filename,
+                report->lineno);
+    }
+
     if ((report->flags & JSREPORT_WARNING) != 0) {
         /* We manually insert "WARNING" into the output instead of
          * having GJS_DEBUG_WARNING because it's convenient to
@@ -929,3 +1006,34 @@ gjs_get_type_name(jsval value)
         return "<unknown>";
     }
 }
+
+jsval
+gjs_date_from_time_t (JSContext *context, time_t time)
+{
+    JSObject *date;
+    JSClass *date_class;
+    JSObject *date_constructor;
+    jsval date_prototype;
+    jsval args[1];
+
+    if (!JS_EnterLocalRootScope(context))
+        return JSVAL_VOID;
+
+    if (!JS_GetClassObject(context, JS_GetGlobalObject(context), JSProto_Date,
+                           &date_constructor))
+        gjs_fatal("Failed to lookup Date prototype");
+
+    if (!JS_GetProperty(context, date_constructor, "prototype", &date_prototype))
+        gjs_fatal("Failed to get prototype from Date constructor");
+
+    date_class = JS_GetClass(context, JSVAL_TO_OBJECT (date_prototype));
+
+    if (!JS_NewNumberValue(context, ((double) time) * 1000, &(args[0])))
+        gjs_fatal("Failed to convert time_t to number");
+
+    date = JS_ConstructObjectWithArguments(context, date_class,
+                                           NULL, NULL, 1, args);
+
+    JS_LeaveLocalRootScope(context);
+    return OBJECT_TO_JSVAL(date);
+}
